import dateutil
from datetime import date, datetime, timezone as tz
import pandas as pd

def float_as_currency(val: float) -> str:
    return "${:,.2f}".format(round(val, 2))


def int_tryParse(value) -> int:
    try:
        return int(value)
    except:
        return None


def date_tryParse(value) -> date:
    return force_to_date_or_None(value)

def datestamp_tryParse(value, include_time: bool = True, include_ms: bool = True) -> datetime:
    try:
        val = force_to_datetime_or_None(value)

        if not include_time:
            val = datetime.combine(val, datetime.min.time())

        if not include_ms:
            val = val.replace(microsecond=0)

        return val
    except:
        return None

def uuid_tryparse(val) -> uuid.UUID:
    if isinstance(val, uuid.UUID):
        return val
    elif isinstance(val, str) and val != '':
        return uuid.UUID(val)
    elif isinstance(val, float) and math.isnan(val):
        return None
    else:
        raise TypeError(f"Unhandled UUID value: {val}")

def force_to_date_or_None(val):
    my_datetime = force_to_datetime_or_None(val)
    if my_datetime is None:
        return None
    else:
        return my_datetime.date()

def force_to_datetime_or_None(val, force_to_midnight: bool = False):
    try:
        # parse supported inputs to a datetime
        if isinstance(val, datetime):
            ret = val
        elif isinstance(val, date):
            ret = datetime.combine(val, datetime.min.time())
        elif isinstance(val, pd.Timestamp):
            the_date = val.to_pydatetime().date()
            ret = datetime.combine(the_date, datetime.min.time())
        else:
            val = dateutil.parser.parse(val)
            ret = force_to_datetime_or_None(val)

        # force to midnight if requested
        if force_to_midnight:
            ret = datetime.combine(ret.date(), datetime.min.time())

        # return value that was parsed
        return ret
    except:
        return None


if __name__ == "__main__":
    try_vals = [
        date.today(),
        '10.21.21',
        '10/21/21',
        None,
        datetime.today(),
        '2014-08-01 11:00:00+02:00'
    ]

    results = map(force_to_datetime_or_None, try_vals)

    print(list(results))

    results = map(lambda x: datestamp_tryParse(x, include_ms=False), try_vals)
    print(list(results))


import threading
from typing import Callable, Any, List
import time
import logging

logger = logging.getLogger("cooptools.datarefresher")


class DataRefresher:
    def __init__(self,
                 name: str,
                 refresh_callback: Callable[[], Any],
                 refresh_interval_ms: int,
                 timeout_on_fail_ms: int = None,
                 as_async: bool = False):
        self._name = name
        self._last_refresh = None
        self._refresh_callback = refresh_callback
        self._cached_return = None
        self._start_timeout = None
        self._timeout_on_fail_ms = timeout_on_fail_ms if timeout_on_fail_ms else refresh_interval_ms
        self.refresh_interval_ms = refresh_interval_ms
        self._async = as_async

        # Set initial cache
        self._try_refresh_cache()

        if self._async: self._init_refresh_thread()

    def _init_refresh_thread(self):
        refresh_thread = threading.Thread(target=self._async_thread_loop, daemon=True)
        refresh_thread.start()

    @property
    def in_timeout(self):
        # check if can clear the timeout
        if self._start_timeout and time.perf_counter() - self._start_timeout > self._timeout_on_fail_ms:
            self._start_timeout = None

        return self._start_timeout is not None

    @property
    def latest(self):
        if self._async:
            return self._cached_return
        else:
            return self._check_refresh_latest()

    def refresh(self):
        self._try_refresh_cache()
        return self._cached_return

    def _set_cache(self):
        """ Set cache with value from the refresh callback. Acquires a thread lock before setting cache """
        tic = time.perf_counter()
        return_of_callback = self._refresh_callback()
        with threading.Lock():
            self._cached_return = return_of_callback
            toc = time.perf_counter()
            logger.info(f"cache refreshed for \"{self._name}\" in {round(toc-tic, 3)} sec")

    def _try_refresh_cache(self):
        """ Try the refresh callback. if fail, start the timeout timer """
        try:
            self._set_cache()
            self._last_refresh = time.perf_counter()
        except Exception as e:
            logger.error(e)
            self._start_timeout = time.perf_counter()

    def _check_refresh_latest(self):
        """ Checks the state of the instance and whether or not to refresh"""
        # dont do any refresh during a timeout
        if self.in_timeout:
            pass
        # check if time to refresh cache
        elif self._last_refresh is None or \
                time.perf_counter() - self._last_refresh > self.refresh_interval_ms / 1000:
            self._try_refresh_cache()

        return self._cached_return

    def _async_thread_loop(self):
        while True:
            self._check_refresh_latest()
            time.sleep(0.5)

    @property
    def last_refresh(self):
        return self._last_refresh

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    def takes_some_time():
        time.sleep(2)
        return {"a": time.perf_counter()}

    callback = takes_some_time

    refresh_interval_sec = 5
    df = DataRefresher('key', callback, 5 * 1000, as_async=True)

    while True:
        print(df.latest)
        time.sleep(.5)




import pandas as pd
from typing import List, Tuple
import numpy as np


def pretty_print_dataframe(df: pd.DataFrame, title: str = None):
    pretty_format_dataframe(df, title)

def pretty_format_dataframe(df: pd.DataFrame, title: str = None):
    ret = ""

    if title:
        ret += title

    with pd.option_context('display.max_rows', 500, 'display.max_columns', 2000, 'display.width', 250):
        ret += f"{df}\n"

    return ret


def pretty_print_list_of_list(lst_of_lst: List[List], title: str = None):
    print(pretty_format_list_of_list(lst_of_lst, title))

def pretty_format_list_of_list(lst_of_lst: List[List], title: str = None):
    ret = ""

    if title:
        ret += title

    n = np.array([[len(str(x)) for x in l] for l in lst_of_lst])
    maxxs = np.amax(n, axis=0)
    # Print the rows
    for row in lst_of_lst:
        joined = ''.join(str(x).ljust(maxxs[ii] + 2) for ii, x in enumerate(row))
        ret += f"\n\t{joined}"

    return ret

def pretty_format_list_of_tuple(lst_of_tup: List[Tuple], title: str = None):
    ret = ""

    if title:
        ret += title

    n = np.array([[len(str(x)) for x in l] for l in lst_of_tup])
    maxxs = np.amax(n, axis=0)
    # Print the rows
    for row in lst_of_tup:
        joined = ''.join(str(x).ljust(maxxs[ii] + 2) for ii, x in enumerate(row))
        ret += f"\n\t{joined}"

    return ret


from typing import Dict, Any
import numpy.random as rnd
from coopbugger.buggers import timer

# @timer
def weighted_random_choice(choices_dict: Dict[Any, float], seed: int = None):
    if seed is not None:
        rnd.seed(seed)

    total = sum([v for k, v in choices_dict.items()])
    selection_probs = [v / total for k, v in choices_dict.items()]

    return list(choices_dict)[rnd.choice(len(choices_dict), p=selection_probs)]

if __name__ == "__main__":
    import logging
    logging.basicConfig(level=logging.DEBUG)

    choices = {
        "a": 1,
        "b": 2,
        "c": 1.5,
        "d": 4
    }

    #TEST 1
    # rnd.seed(0)
    # ret = weighted_random_choice(choices)
    # print(ret)

    #TEST 2
    rets = {}
    for ii in range(0, 1000):
        ret = weighted_random_choice(choices)
        rets.setdefault(ret, 0)
        rets[ret] += 1

    for k, v in choices.items():
        print(f"{k} -- {v} ({round(v / sum([x for x in choices.values()]) * 100, 1)}%)")

    for k, v in rets.items():
        print(f"{k} -- {v} ({round(v / sum([x for x in rets.values()]) * 100, 1)}%)")





import functools
from threading import Lock

def synchronized(lock: Lock):
    """ Synchronization decorator """
    def wrap(f):
        @functools.wraps(f)
        def newFunction(*args, **kw):
            with lock:
                return f(*args, **kw)
        return newFunction
    return wrap


from typing import List, Tuple

def plot_series(points: List[Tuple[int, float]],
                 ax,
                 color=None,
                 type=None,
                 label=None,
                 line_width: int = None,
                 line_style: str = None,
                 point_size = None):
    """

    :param points:
    :param ax:
    :param color:
    :param type:
    :param label:
    :param line_width:
    :param line_style: [‘solid’, ‘dashed’, ‘dashdot’, ‘dotted’, (offset, on-off-dash-seq), '-', '--', '-.', ':', 'None', ' ', '']
    :return:
    """

    if points is None or len(points) == 0:
        return

    res = list(zip(*points))

    if type is None:
        ax.plot(res[0], res[1], color=color, linewidth=line_width, label=label, linestyle=line_style)
    elif type == 'scatter':
        ax.scatter(res[0], res[1], color=color, label=label, s=point_size)


from cooptools.plotting import plot_series
import cooptools.statistics.controlChart.controlChart as cc
import matplotlib.pyplot as plt
from typing import List

def animate_control(data, ax, trailing_window=None, x_s: List = None):
    control = cc.control_data_and_deltas([float(x) for x in data], trailing_window=trailing_window)
    ax.clear()
    plot_control_chart(control, ax, trailing_window, x_s=x_s)


def plot_show_and_save(control_data, trailing_window: int, show: bool = True, x_s: List = None, out_dirs: List = None):
    f, axes = plt.subplots(1, 1, figsize=(15, 10))
    plot_control_chart(control_data, axes, trailing_window=trailing_window, x_s=x_s)

    if show:
        plt.show()

    if out_dirs is not None:
        for dir in out_dirs:
            f.savefig(dir)


def plot_control_chart(control, ax, trailing_window=None, x_s: List = None):
    if control is None or ax is None:
        return

    shade_ooc(control, ax, x_s=x_s)
    plot_lcl_trend(control, ax, x_s=x_s)
    plot_mean_trend(control, ax, x_s=x_s)
    plot_ucl_trend(control, ax, x_s=x_s)
    plot_ucl(control, ax, trailing_window, x_s=x_s)
    plot_p_one_stdev(control, ax, trailing_window, x_s=x_s)
    plot_p_two_stdev(control, ax, trailing_window, x_s=x_s)
    plot_m_one_stdev(control, ax, trailing_window, x_s=x_s)
    plot_m_two_stdev(control, ax, trailing_window, x_s=x_s)
    plot_lcl(control, ax, trailing_window, x_s=x_s)
    plot_mean(control, ax, trailing_window, x_s=x_s)
    plot_data(control, ax, x_s=x_s)
    plot_out_of_controls(control, ax, x_s=x_s)
    plot_outliers(control, ax, x_s=x_s)


def shade_ooc(control, ax, x_s: List = None):
    ax.fill_between(x_s if x_s else range(len(control)),
                    min([x['data_point'] for x in control] + [x['lcl'] for x in control if x['lcl']]),
                    max([x['data_point'] for x in control] + [x['ucl'] for x in control if x['ucl']]),
                    where=([x['out_of_control'] or control[min(len(control) - 1, ii + 1)]['out_of_control'] for ii, x in
                            enumerate(control)]),
                    alpha=0.25)


def plot_line(control, ax, key, color, x_s: List = None, trailing_window=None, last_only: bool = False, line_style='--',
              line_width=2):
    if last_only:
        data = [(x_s[ind] if x_s else ind, control[-1][key]) for ind, x in enumerate(control) if
                x[key] is not None and ind > len(control) - trailing_window]
    else:
        data = [(x_s[ind] if x_s else ind, x[key]) for ind, x in enumerate(control) if x[key] is not None]
    plot_series(data, ax, color=color, line_style=line_style, line_width=line_width)


def plot_ucl(control, ax, trailing_window, x_s: List = None):
    plot_line(control,
              ax,
              x_s=x_s,
              trailing_window=trailing_window,
              last_only=True,
              key='ucl',
              color='r',
              line_style='--',
              line_width=2)


def plot_p_one_stdev(control, ax, trailing_window, x_s: List = None):
    plot_line(control,
              ax,
              x_s=x_s,
              trailing_window=trailing_window,
              last_only=True,
              key='p_one_stdev',
              color='grey',
              line_style='--',
              line_width=1)


def plot_p_two_stdev(control, ax, trailing_window, x_s: List = None):
    plot_line(control,
              ax,
              x_s=x_s,
              trailing_window=trailing_window,
              last_only=True,
              key='p_two_stdev',
              color='grey',
              line_style='--',
              line_width=1)


def plot_m_one_stdev(control, ax, trailing_window, x_s: List = None):
    plot_line(control,
              ax,
              x_s=x_s,
              trailing_window=trailing_window,
              last_only=True,
              key='m_one_stdev',
              color='grey',
              line_style='--',
              line_width=1)


def plot_m_two_stdev(control, ax, trailing_window, x_s: List = None):
    plot_line(control,
              ax,
              x_s=x_s,
              trailing_window=trailing_window,
              last_only=True,
              key='m_two_stdev',
              color='grey',
              line_style='--',
              line_width=1)


def plot_lcl(control, ax, trailing_window, x_s: List = None):
    plot_line(control,
              ax,
              x_s=x_s,
              trailing_window=trailing_window,
              last_only=True,
              key='lcl',
              color='r',
              line_style='--',
              line_width=2)


def plot_ucl_trend(control, ax, x_s: List = None):
    plot_line(control,
              ax,
              x_s=x_s,
              last_only=False,
              key='ucl',
              color='r',
              line_style='--',
              line_width=1)


def plot_lcl_trend(control, ax, x_s: List = None):
    plot_line(control,
              ax,
              x_s=x_s,
              last_only=False,
              key='lcl',
              color='r',
              line_style='--',
              line_width=1)


def plot_mean(control, ax, trailing_window, x_s: List = None):
    plot_line(control,
              ax,
              x_s=x_s,
              trailing_window=trailing_window,
              last_only=True,
              key='mean',
              color='g',
              line_style='--',
              line_width=2)


def plot_mean_trend(control, ax, x_s: List = None):
    plot_line(control,
              ax,
              x_s=x_s,
              last_only=False,
              key='mean',
              color='g',
              line_style='--',
              line_width=1)


def plot_data(control, ax, x_s: List = None):
    data = [(x_s[ind] if x_s else ind, x['data_point']) for ind, x in enumerate(control) if x['data_point'] is not None]
    plot_series(data, ax, color='grey', type='scatter', point_size=2)


def list_from_lists(lists_of_lists: List[List]):
    the_set = set()

    for alist in lists_of_lists:
        if alist is not None:
            try:
                the_set.update(list_from_lists(alist))
            except Exception as e:
                the_set.update(alist)
    return list(the_set)


def plot_out_of_controls(control, ax, x_s: List = None):
    triggering_indexes = list_from_lists(
        [x['out_of_control_triggering_points'] for x in control if x['out_of_control_triggering_points']])
    oocs = [(x_s[ind] if x_s else ind, control[ind]['data_point']) for ind in triggering_indexes]
    plot_series(oocs, ax, color='y', type='scatter', point_size=6)


def plot_outliers(control, ax, x_s: List = None):
    outliers = set()
    for ii, x in enumerate(control):
        if x['out_of_limit_high']:
            outliers.update([(x_s[ind] if x_s else ind, control[ind]['data_point']) for ind in x['out_of_limit_high']])
        if x['out_of_limit_low']:
            outliers.update([(x_s[ind] if x_s else ind, control[ind]['data_point']) for ind in x['out_of_limit_low']])

    if len(outliers) > 0:
        plot_series(list(outliers), ax, color='r', type='scatter')


if __name__ == "__main__":
    import random as rnd
    import pandas as pd
    import cooptools.pandasHelpers as ph

    data = [rnd.normalvariate(100, 25) for ii in range(250)]
    # cc = pd.DataFrame(control_data_and_deltas(data))
    trailing_window = 100
    control_data = cc.control(data, trailing_window=trailing_window)
    cc_df = pd.DataFrame(control_data)
    ph.pretty_print_dataframe(cc_df)

    f, axes = plt.subplots(1, 1, figsize=(15, 10))
    plot_control_chart(control_data, axes, trailing_window=trailing_window)
    plt.show()



from typing import Tuple, List
import math

def line_intersection(line1: Tuple[Tuple[float, float]], line2: Tuple[Tuple[float, float]]):
    xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(xdiff, ydiff)
    if div == 0:
        raise Exception('lines do not intersect')

    d = (det(*line1), det(*line2))
    x = det(d, xdiff) / div
    y = det(d, ydiff) / div
    return x, y

def point_in_polygon(point: Tuple[float, float], poly: List[Tuple[float, float]]):
    # https://en.wikipedia.org/wiki/Point_in_polygon#:~:text=One%20simple%20way%20of%20finding,an%20even%20number%20of%20times.
    # https://en.wikipedia.org/wiki/Even%E2%80%93odd_rule


    # Determine if the point is in the polygon.
    #
    # Args:
    #   x -- The x coordinates of point.
    #   y -- The y coordinates of point.
    #   poly -- a list of tuples [(x, y), (x, y), ...]
    #
    # Returns:
    #   True if the point is in the path or is a corner or on the boundary

    num = len(poly)
    j = num - 1
    c = False
    for i in range(num):
        if (point[0] == poly[i][0]) and (point[1] == poly[i][1]):
            # point is a corner
            return True
        if ((poly[i][1] > point[1]) != (poly[j][1] > point[1])):
            slope = (point[0]- poly[i][0]) * (poly[j][1] - poly[i][1]) - (poly[j][0] - poly[i][0]) * (point[1] - poly[i][1])
            if slope == 0:
                # point is on boundary
                return True
            if (slope < 0) != (poly[j][1] < poly[i][1]):
                c = not c
        j = i
    return c

def collinear_points(points: List[Tuple[float, float]]):

    if len(points) < 3:
        return True

    for ii in range(2, len(points)):
        a = points[ii - 2]
        b = points[ii - 1]
        c = points[ii]
        tri_area = a[0] * (b[1] - c[1]) + b[0] * (c[1] - a[1]) + c[0] * (a[1] - b[1])
        if math.isclose(tri_area, 0):
            continue
        else:
            return False

    return True

if __name__ == "__main__":
    #TEST1
    poly = [
        (2, 2),
        (3, 3),
        (2, 3),
        (3, 2)
    ]

    point = (1, 1)
    print(point_in_polygon(point, poly))
    point = (2.5, 2.5)
    print(point_in_polygon(point, poly))

    #TEST2
    a = (0, 1)
    b = (0, 2)
    c = (0, 3)
    d = (0, 4)
    e = (0, 5)
    f = (0, 6)
    print(collinear_points([a, b, c, d, e, f]))

    #TEST3
    a = (0, 1.000000001)
    b = (0, 1.000000002)
    c = (0, 3)
    print(collinear_points([a, b, c]))





def strictly_increasing(L):
    return all(x<y for x, y in zip(L, L[1:]))

def strictly_decreasing(L):
    return all(x>y for x, y in zip(L, L[1:]))

def alternating(L):
    return all(x>0 != y>0 for x, y in zip(L, L[1:]))

def non_increasing(L):
    return all(x>=y for x, y in zip(L, L[1:]))

def non_decreasing(L):
    return all(x<=y for x, y in zip(L, L[1:]))

def monotonic(L):
    return non_increasing(L) or non_decreasing(L)


import pandas as pd
from typing import Dict, Callable, List
import datetime
from cooptools.printing import pretty_format_list_of_tuple, pretty_format_dataframe

class PandasMissingColumnsException(Exception):
    ''' Raised when a user tries to create an account that already exists'''
    def __init__(self, columns_present: List[str], required_columns: List[str]):
        self.columns_present = columns_present
        self.required_columns = required_columns
        self.missing_columns = list(set(required_columns) - set(columns_present))
        self.message = f"All columns [{self.required_columns}] required but columns [{self.missing_columns}] were missing"
        super().__init__(self.message)


class PandasFillColumnTypeException(Exception):
    def __init__(self, str):
        Exception().__init__(str)

    pass



def pretty_print_dataframe(df: pd.DataFrame,
                           title: str = None,
                           max_columns: int = 2000,
                           max_rows: int = 500,
                           display_width: int = 250
                            ):
    ret = ""

    if title:
        ret += title
        print(title)

    with pd.option_context('display.max_rows', max_rows, 'display.max_columns', max_columns, 'display.width', display_width):
        ret += f"{df}\n"
        print(f"{df}\n")

    return ret

def find_the_type_conversion_errors_and_raise(df: pd.DataFrame, column_type_definition: Dict):

    errors = []
    for column, requested_type in column_type_definition.items():
        try:
            df[column].astype(requested_type)
        except:
            for index, row in df.iterrows():
                try:
                    row[column].astype(requested_type)
                except:
                    errors.append((index, column, row[column], requested_type, type(row[column])))

    errors_df = pd.DataFrame(errors,
                             columns=['row_index', 'column_name', 'value', 'requested_type_conversion', 'current_type'])

    raise ValueError(f"Error performing the following conversions: \n{pretty_format_dataframe(errors_df)}")

def clean_currency(x):
    """ If the value is a string, then remove currency symbol and delimiters
    otherwise, the value is numeric and can be converted
    """
    if isinstance(x, str):
        return(x.replace('$', '').replace(',', '').replace('(', '-').replace(')', ''))
    return(x)


def convert_pandas_data_columns_to_type(df: pd.DataFrame, column_type_definition: Dict) -> pd.DataFrame:

    if df is None:
        return None


    # handle currency conversions for floats/ints:
    columns_that_may_need_cleaned = [k for k, v in column_type_definition.items() if
             v in (float, int) and k in df.columns]
    for column in columns_that_may_need_cleaned:
        pd.options.mode.chained_assignment = None
        df[column] = df[column].map(clean_currency)
        pd.options.mode.chained_assignment = 'warn'

    # Cast columns as type (excluding dates and ints)
    types = {k:v for k, v in column_type_definition.items() if v not in (datetime.date, datetime.datetime) and k in df.columns}
    try:
        df = df.astype(types)
    except:
        find_the_type_conversion_errors_and_raise(df[[x for x in types.keys()]], types)

    # handle date conversions
    for col, type in {k: v for k, v in column_type_definition.items() if v in (datetime.date, datetime.datetime) and k in df.columns}.items():
        df[col] = pd.to_datetime(df[col])
        if type == datetime.date:
            df[col] = df[col].dt.normalize()

    return df


def verify_all_required_columns(df: pd.DataFrame, required_columns: List[str]):
    if not all(column in df.columns for column in required_columns):
        raise PandasMissingColumnsException(columns_present=[col for col in df.columns], required_columns=required_columns)

def clean_a_dataframe(df: pd.DataFrame,
                      column_type_definition: Dict,
                      case_sensitive: bool = False,
                      allow_partial_columnset:bool=False,
                      fill_missing: bool=False,
                      column_name_replacement: Callable[[pd.DataFrame], Dict] = None,
                      keep_extra: bool = False) -> pd.DataFrame:

    required_columns = [key for key, value in column_type_definition.items()]

    # update based on case-sensitivity
    if not case_sensitive:
        df.columns = map(str.lower, df.columns)
        col_indexes = {col: df.columns.get_loc(col.lower()) if col.lower() in df.columns else None for col in required_columns}
        name_replacements = {[x for x in df.columns][ind]: col for col, ind in col_indexes.items() if ind is not None}
        df.rename(columns=name_replacements, inplace=True)

    # map columns that dont match name
    if column_name_replacement is not None:
        df = df.rename(columns=column_name_replacement())

    # Filter columns
    if not keep_extra:
        df = df[[col for col in required_columns if col in df.columns]]

    # raise error if not all columns exist
    if not allow_partial_columnset:
        verify_all_required_columns(df, required_columns)

    # handle empty dataframe
    if not any(df):
        df = pd.DataFrame(columns=required_columns)

    # add missing columns
    missing_columns = [column for column in required_columns if column not in df.columns]
    if fill_missing:
        for col in missing_columns:
            requested_type = column_type_definition[col]
            if requested_type in [int]:
                raise PandasFillColumnTypeException(f"Unable to create an empty column of type {requested_type}. {requested_type} is not nullable.")
            df[col] = pd.Series([], dtype=requested_type)

    # column type conversions
    df = convert_pandas_data_columns_to_type(df, column_type_definition)

    # filter empty rows
    df.replace({'nan': None}, inplace=True)
    df.dropna(how='all', inplace=True)

    # return
    return df

def try_parse_and_localize_date_series(date_column):
    try:
        return pd.to_datetime(date_column).dt.tz_localize(None)
    except:
        return pd.to_datetime(date_column)

def import_data_from_csv_with_specified_columns(filepath: str,
                                                column_type_definition: Dict,
                                                allow_partial:bool=False,
                                                fill_missing: bool=False) -> pd.DataFrame:
    # Get file contents
    if filepath is None or filepath == '':
        return None
    df = pd.read_csv(filepath)
    if df is None:
        return None

    # Return
    return clean_a_dataframe(df, column_type_definition=column_type_definition, allow_partial_columnset=allow_partial, fill_missing=fill_missing)

if __name__ == "__main__":
    import numpy as np

    df = pd.read_csv('../tests/testdata/dummy_data_clean.csv')
    column_definitions = {"my_clean_int": int,
                          "my_clean_str": str,
                          "my_clean_date": datetime.date,
                          "my_missing_column": np.int64}

    converted = clean_a_dataframe(df, column_type_definition=column_definitions, allow_partial_columnset=True)

    converted['newdate'] = converted['my_clean_date'] + datetime.timedelta(days=1)

    print(converted.dtypes)



    print(converted.dtypes['my_clean_date'])
    print(converted.dtypes['my_clean_date'] == np.datetime64)
    print(converted)


import os
import shutil
import logging
import json

LOCALAPP_ROOT = 'LOCALAPPDATA'


def check_and_make_new_proj_localapp(app_root, proj_name):
    root = check_and_make_localapp_application_path_dir(app_root)
    return check_and_make_proj_path_dir(root, proj_name)


def root_and_name(root, name):
    return f"{root}\\{name}"


def check_and_make_proj_path_dir(root, proj_name):
    proj_dir = root_and_name(root, proj_name)
    check_and_make_dir(proj_dir)

    return proj_dir


def localapp_root(app_name):
    local_app_data = os.getenv(LOCALAPP_ROOT)
    return f"{local_app_data}\\{app_name}"


def check_and_make_localapp_application_path_dir(application_root):
    la_root = localapp_root(application_root)
    check_and_make_dir(la_root)
    return la_root


def check_and_make_dir(dir):
    chk = os.path.isdir(dir)
    if not chk:
        os.makedirs(dir)


def copy_file_from_to(filepath_to_copy: str, to_filepath: str):
    shutil.copy2(filepath_to_copy, to_filepath)


def try_save_json_data(my_jsonable_data, file_path):
    with open(file_path, 'w') as outfile:
        outfile.write(json.dumps(my_jsonable_data, indent=4))


def try_load_json_data(file_path):
    ret = None

    if not os.path.isfile(file_path):
        return None

    with open(file_path, 'r+') as outfile:
        try:
            ret = json.load(outfile)
        except Exception as e:
            logging.error(f"Unable to read file: {file_path}: {e}")
    return ret


def save_application_json(my_jsonable_data, app_root, filename: str):
    app_dir = check_and_make_localapp_application_path_dir(app_root)
    filename = filename.replace(".json", "").replace(".JSON", "").replace(".Json", "")
    return try_save_json_data(my_jsonable_data, f"{app_dir}\\{filename}.json")


def save_project_json(my_jsonable_data, app_root, project_name, filename: str):
    proj_dir = check_and_make_new_proj_localapp(app_root, project_name)
    filename = filename.replace(".json", "").replace(".JSON", "").replace(".Json", "")
    return try_save_json_data(my_jsonable_data, f"{proj_dir}\\{filename}.json")


def load_application_json(app_root, filename):
    app_dir = check_and_make_localapp_application_path_dir(app_root)
    file_path = f"{app_dir}\\{filename}.json"
    return try_load_json_data(file_path)


def load_project_json(app_root, project, filename):
    proj_dir = check_and_make_new_proj_localapp(app_root, project)
    file_path = f"{proj_dir}\\{filename}.json"
    return try_load_json_data(file_path)


from configparser import ConfigParser
import os
import logging
from typing import Dict, Any

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

class ConfigStateException(Exception):
    def __init__(self, txt: str = None):
        logger.error(txt)
        super().__init__(txt)

class ConfigStateFactory:

    def __init__(self,
                 config_definition_dict: Dict[str, Any],
                 config_file_path: str = None,
                 config_file_header:str = None):

        if config_definition_dict is None:
            raise ConfigStateException(f"config_definition_dict cannot be None")

        if config_file_path is not None and not os.path.isfile(config_file_path):
            raise ConfigStateException(f"config_file_path {config_file_path} does not exist.")

        self.parser = ConfigParser()
        self.current_file_path = config_file_path
        self.current_file_header = config_file_header
        self._config_definition_dict = config_definition_dict

        self.loaded = False

        self._q = type('accessor', (object,), self._config_definition_dict)

        self._try_load_pass(config_file_path, config_file_header)

    def __getitem__(self, item):
        return self.get_config(item)

    def _try_load_pass(self, config_file_path:str, config_file_header: str):
        try:
            self.load_from_file(config_file_path=config_file_path, config_file_header=config_file_header)
        except:
            pass
        finally:
            # set attributes
            for k, v in self._config_definition_dict.items():
                setattr(self, k, self._try_access_parser_default(config_file_header, k, v))

    def _config_file_exists(self):
        return os.path.isfile(self.current_file_path)

    def load_from_file(self, config_file_path: str = None, config_file_header: str = None):
        if config_file_path is not None:
            self.current_file_path = config_file_path

        if config_file_header is not None:
            self.current_file_header = config_file_header

        if not self._config_file_exists():
            issue = f"Unable to load config from directory: \"{self.current_file_path}\" does not exist"
            raise ConfigStateException(issue)

        try:
            self.parser.read(self.current_file_path)
            logger.info(f"Config set from directory: {self.current_file_path}")
            self.loaded = True
        except Exception as e:
            issue = f"Unable to load config from directory: {self.current_file_path}" \
                    f"\n{e}"
            raise ConfigStateException(issue) from e

    def _try_access_parser_default(self, header, name, default):
        if self.loaded:
            try:
                return self.parser.get(header, name)
            except Exception as e:
                raise ConfigStateException(str(e)) from e
        else:
            try:
                return default()
            except:
                return default

    def _get_config(self, header, name):
        if name not in self._config_definition_dict.keys():
            raise ConfigStateException(f"Requested config [\"{name}\"] was not defined")
        default = self._config_definition_dict[name]
        val = self._try_access_parser_default(header, name, default)
        logger.debug(f"Value for {header}|{name}: {val}")
        return val

    def get_config(self, name):
        return self._get_config(self.current_file_header, name)

    @property
    def config_dir(self):
        return os.path.dirname(self.current_file_path)

    @property
    def config_file_path(self):
        return self.current_file_path


if __name__ == "__main__":
    my_dict = {"a": None,
               "b": 2,
               "c": "TRYIT",
               "d": lambda: 123}

    con = ConfigStateFactory(my_dict)

    print(con.get_config("a"), con.get_config("b"), con.get_config("c"))

    print(con.a, con.b, con.c, con.d)
    print(con["b"])




from typing import Dict, List, Union, Any, Callable
import numpy as np
from cooptools.statistics.trends import *
import math

def n_of_x_outside_mean_and_stdev(data,
                                  start,
                                  stop,
                                  mean,
                                  stdev,
                                  n_stdevs,
                                  x,
                                  n_of,
                                  high_low_both: int = None,
                                  outside: bool = True):

    if math.isclose(high_low_both, 0) or high_low_both is None:
        compare = lambda k: k < mean - n_stdevs * stdev or k > mean + n_stdevs * stdev
    elif high_low_both < 0: compare = lambda k: k < mean - n_stdevs * stdev
    elif high_low_both > 0: compare = lambda k: k > mean + n_stdevs * stdev
    else: raise ValueError(f"Unable to handle high_low_both value {high_low_both}")

    if not outside:
        _compare = lambda x: not compare(x)
    else:
        _compare = compare

    return qualify_data_points(data, start, stop, x, n_of, _compare)


def qualify_data_points(data,
                        start,
                        stop,
                        x,
                        n_of,
                        compare: Callable[[Union[float, int]], bool] = None):
    rule_triggers = []

    for ii in range(start + x - 1, stop):
        qualifiers = [x for x in data[ii - x + 1: ii + 1] if compare(x)]
        if len(qualifiers) >= n_of:
            rule_triggers.append([jj for jj in range(ii - x + 1, ii + 1)])

    return rule_triggers

def qualify_data_sequence(data,
                       start,
                       stop,
                       x,
                       compare: Callable[[List[Union[float, int]]], bool] = None):
    rule_triggers = []

    for ii in range(start + x - 1, stop):
        qualifiers = data[ii - x + 1: ii + 1] if compare(data[ii - x + 1: ii + 1]) else None
        if qualifiers is not None:
            rule_triggers.append([jj for jj in range(ii - x + 1, ii + 1)])

    return rule_triggers


def list_from_lists(lists_of_lists: List[List]):
    the_set = set()

    for alist in lists_of_lists:
        if alist is not None:
            try:
                the_set.update(list_from_lists(alist))
            except Exception as e:
                the_set.update(alist)
    return list(the_set)


def control(data: List[Union[int, float]],
            set_baseline_periods: int = None,
            trailing_window: int = None) -> List[Dict[str, Any]]:
    """ Takes a set of data and determines if it statistically in control

    :param: data: List of integers or floats that represents the data set to evaluate
    :param: set_baseline_periods: The number of periods that must pass before evaluating the statistics
    :param: trailing_window: The amount of past periods to evaluate for control (should be at least 15 since control includes
     checks of at least past 15 periods
    :return: List Dict of values by period
    """

    # init optional params
    if set_baseline_periods is None:
        set_baseline_periods = 20

    if trailing_window is None:
        trailing_window = 20

    if trailing_window <= 0:
        raise ValueError(f"The trailing window must be greater than zero. Ideally, this value is at least 20")

    # init the tracking dicts
    deltas = {}
    stdevs = {}
    means = {}
    p_one_stdev = {}
    p_two_stdev = {}
    m_one_stdev = {}
    m_two_stdev = {}
    lcls = {}
    ucls = {}
    out_of_limit_high = {}
    out_of_limit_low = {}
    two_of_three_high_2stdev = {}
    two_of_three_low_2stdev = {}
    four_of_five_high_1stdev = {}
    four_of_five_low_1stdev = {}
    seven_high = {}
    seven_low = {}
    seven_trend_high = {}
    seven_trend_low = {}
    eight_out_of_1stdev = {}
    fifteen_within_1stdev = {}
    fourteen_alternate = {}
    out_of_control_triggering_points = {}
    out_of_control = {}

    # iterate the data
    for ii in range(0, len(data)):
        # track deltas
        deltas[ii] = data[ii] - data[ii-1] if ii > 0 else None

        # track means and stdevs
        start_period = max(0, ii - trailing_window - 1)
        means[ii] = round(np.mean(data[start_period: ii]), 3) if ii >= set_baseline_periods - 1 else None
        stdevs[ii] = round(np.std(data[start_period:ii]), 3) if ii >= set_baseline_periods - 1 else None

        # track LCL and UCL, and stdev lvls
        lcls[ii] = (means[ii] - 3 * stdevs[ii]) if ii >= set_baseline_periods - 1 else None
        ucls[ii] = means[ii] + 3 * stdevs[ii] if ii >= set_baseline_periods - 1 else None
        p_one_stdev[ii] = means[ii] + stdevs[ii] if ii >= set_baseline_periods - 1 else None
        p_two_stdev[ii] = means[ii] + 2 * stdevs[ii] if ii >= set_baseline_periods - 1 else None
        m_one_stdev[ii] = means[ii] - stdevs[ii] if ii >= set_baseline_periods - 1 else None
        m_two_stdev[ii] = means[ii] - 2 * stdevs[ii] if ii >= set_baseline_periods - 1 else None

        # Track control rules
        out_of_limit_high[ii] = [x + start_period for x, n in enumerate(data[start_period:ii + 1]) if n > ucls[ii]] if ii >= set_baseline_periods - 1 else None
        out_of_limit_low[ii] = [x + start_period for x, n in enumerate(data[start_period:ii + 1]) if n < lcls[ii]] if ii >= set_baseline_periods - 1 else None
        two_of_three_high_2stdev[ii] = n_of_x_outside_mean_and_stdev(data, start_period, ii, means[ii], stdevs[ii], 2, 3, 2, 1) if ii >= set_baseline_periods - 1 else None
        two_of_three_low_2stdev[ii] = n_of_x_outside_mean_and_stdev(data, start_period, ii, means[ii], stdevs[ii], 2, 3, 2, -1) if ii >= set_baseline_periods - 1 else None
        four_of_five_high_1stdev[ii] = n_of_x_outside_mean_and_stdev(data, start_period, ii, means[ii], stdevs[ii], 1, 5, 4, 1) if ii >= set_baseline_periods - 1 else None
        four_of_five_low_1stdev[ii] = n_of_x_outside_mean_and_stdev(data, start_period, ii, means[ii], stdevs[ii], 1, 5, 4, -1) if ii >= set_baseline_periods - 1 else None
        seven_high[ii] = n_of_x_outside_mean_and_stdev(data, start_period, ii, means[ii], stdevs[ii], 0, 7, 7, 1) if ii >= set_baseline_periods - 1 else None
        seven_low[ii] = n_of_x_outside_mean_and_stdev(data, start_period, ii, means[ii], stdevs[ii], 0, 7, 7, -1) if ii >= set_baseline_periods - 1 else None
        seven_trend_high[ii] = qualify_data_sequence(data, max(ii - 7, start_period), ii + 1, 7, strictly_increasing) if ii >= set_baseline_periods - 1 else None
        seven_trend_low[ii] = qualify_data_sequence(data, max(ii - 7, start_period), ii + 1, 7, strictly_decreasing) if ii >= set_baseline_periods - 1 else None
        eight_out_of_1stdev[ii] = n_of_x_outside_mean_and_stdev(data, start_period, ii, means[ii], stdevs[ii], 1, 8, 8, 0) if ii >= set_baseline_periods - 1 else None
        fifteen_within_1stdev[ii] = n_of_x_outside_mean_and_stdev(data, start_period, ii, means[ii], stdevs[ii], 0, 15, 15, 0, False) if ii >= set_baseline_periods - 1 else None
        fourteen_alternate[ii] = qualify_data_sequence(data, max(ii - 7, start_period), ii + 1, 14, alternating) if ii >= set_baseline_periods - 1 else None


        triggering_points = list_from_lists([
            out_of_limit_high[ii],
            out_of_limit_low[ii],
            two_of_three_high_2stdev[ii],
            two_of_three_low_2stdev[ii],
            four_of_five_high_1stdev[ii],
            four_of_five_low_1stdev[ii],
            seven_high[ii],
            seven_low[ii],
            seven_trend_high[ii],
            seven_trend_low[ii],
            eight_out_of_1stdev[ii],
            fifteen_within_1stdev[ii],
            fourteen_alternate[ii]
        ])

        out_of_control_triggering_points[ii] = triggering_points
        out_of_control[ii] = len(out_of_control_triggering_points[ii]) > 0 if ii >= set_baseline_periods - 1 else None

        # out_of_limit_high[ii] = len([x for x in data[start_period:ii + 1] if x > ucls[ii]]) > 0 if ii > set_baseline_periods else None
        # out_of_limit_low[ii] = len([x for x in data[start_period:ii + 1] if x < lcls[ii]]) > 0 if ii > set_baseline_periods else None
        # two_of_three_high_2stdev[ii] = len([x for x in data[max(ii - 3, start_period):ii + 1] if x > (means[ii] + 2 * stdevs[ii])]) >= 2 if ii > set_baseline_periods else None
        # two_of_three_low_2stdev[ii] = len([x for x in data[max(ii - 3, start_period):ii + 1] if x < (means[ii] - 2 * stdevs[ii])]) >= 2 if ii > set_baseline_periods else None
        # four_of_five_high_1stdev[ii] = len([x for x in data[max(ii - 5, start_period):ii + 1] if x > (means[ii] + stdevs[ii])]) >= 4 if ii > set_baseline_periods else None
        # four_of_five_low_1stdev[ii] = len([x for x in data[max(ii - 5, start_period):ii + 1] if x < (means[ii] - stdevs[ii])]) >= 4 if ii > set_baseline_periods else None
        # seven_high[ii] = len([x for x in data[max(ii - 7, start_period):ii + 1] if x > means[ii]]) == 7 if ii > set_baseline_periods else None
        # seven_low[ii] = len([x for x in data[max(ii - 7, start_period):ii + 1] if x < means[ii]]) == 7 if ii > set_baseline_periods else None
        # seven_trend_high[ii] = strictly_increasing(data[max(ii - 7, start_period):ii + 1]) if ii > set_baseline_periods else None
        # seven_trend_low[ii] = strictly_decreasing(data[max(ii - 7, start_period):ii + 1]) if ii > set_baseline_periods else None
        # eight_out_of_1stdev[ii] = len([x for x in data[max(ii - 8, start_period):ii + 1] if x < means[ii] - stdevs[ii] or x > means[ii] + stdevs[ii]]) == 8 if ii > set_baseline_periods else None
        # fifteen_within_1stdev[ii] = len([x for x in data[max(ii - 15, start_period):ii + 1] if x < means[ii] + stdevs[ii] and x > means[ii] - stdevs[ii]]) == 15 if ii > set_baseline_periods else None
        # fourteen_alternate[ii] = alternating([x for x in list(deltas.values())[max(ii - 14, start_period):ii + 1] if x]) if ii > set_baseline_periods else None


    # format and return
    data = [{'index': x,
             'data_point': data[x],
             'delta': deltas[x],
             'mean': means[x],
             'stdev': stdevs[x],
             'lcl': lcls[x],
             'm_two_stdev': m_two_stdev[x],
             'm_one_stdev': m_one_stdev[x],
             'p_one_stdev': p_one_stdev[x],
             'p_two_stdev': p_two_stdev[x],
             'ucl': ucls[x],
             'out_of_limit_high': out_of_limit_high[x],
             'out_of_limit_low': out_of_limit_low[x],
             'two_of_three_high_2stdev': two_of_three_high_2stdev[x],
             'two_of_three_low_2stdev': two_of_three_low_2stdev[x],
             'four_of_five_high_1stdev': four_of_five_high_1stdev[x],
             'four_of_five_low_1stdev': four_of_five_low_1stdev[x],
             'seven_high': seven_high[x],
             'seven_low': seven_low[x],
             'seven_trend_high': seven_trend_high[x],
             'seven_trend_low': seven_trend_low[x],
             'eight_out_of_1stdev': eight_out_of_1stdev[x],
             'fifteen_within_1stdev': fifteen_within_1stdev[x],
             'fourteen_alternate': fourteen_alternate[x],
             'out_of_control_triggering_points': out_of_control_triggering_points[x],
             'out_of_control': out_of_control[x]} for x in range(len(data))]
    return data

def control_data_and_deltas(data: List[Union[int, float]],
                            set_baseline_periods: int = None,
                            trailing_window: int = None
                            ) -> List[Dict[str, Any]]:
    """
    Calculates the control of both raw data and the deltas of that data to check for overall control

    :param data: raw data to be analyzed
    :param: set_baseline_periods: The number of periods that must pass before evaluating the statistics
    :param: trailing_window: The amount of past periods to evaluate for control (should be at least 15 since control includes
     checks of at least past 15 periods
    :return: List Dict of values by period
    """
    # gather control of raw data
    controlled_data = control(data, set_baseline_periods, trailing_window)

    # gather control of deltas
    deltas = [x['delta'] for x in controlled_data][1:]
    controlled_deltas = control(deltas, set_baseline_periods, trailing_window)
    controlled_deltas.insert(0, {x: None for x in controlled_deltas[0].keys()})

    # merge and return
    ret = []
    for ii in range(len(controlled_data)):
        controlled_deltas[ii] = {f'delta_{key}': value for key, value in controlled_deltas[ii].items()}
        ret.append({**controlled_data[ii], **controlled_deltas[ii]})

    return ret



if __name__ == "__main__":
    import random as rnd
    import pandas as pd
    import cooptools.pandasHelpers as ph
    data = [rnd.normalvariate(100, 25) for ii in range(250)]
    # cc = pd.DataFrame(control_data_and_deltas(data))

    cc = pd.DataFrame(control(data))
    ph.pretty_print_dataframe(cc)





